from . import tools
from . import formula
from . import linear
from . import testing

from .formula import (
    Factor, Term, Formula, Real, Categ, Demean, Binned, Custom,
    robust_eval, factor, design_matrix, design_matrices,
    O, I, R, C, D, B, C0, B0, Drop, sum0 as sum, drop_type
)
from .linear import ols
from .testing import dataset

NONE = Drop.NONE
FIRST = Drop.FIRST
VALUE = Drop.VALUE

try:
    from .general import (
        glm, logit, poisson, negbin, zinf_poisson, zinf_negbin, gols
    )
except:
    pass



##
## formula test suite
##

import re
import pytest
import fastreg as fr
from fastreg import O, I, R, C, FIRST, NONE, VALUE, Factor, Term, Formula, drop_type

def test_factor():
    assert I == I
    assert I != R.x

    assert R.x == R.x
    assert R.x != R.y

def test_term():
    assert I*R.x == R.x
    assert I*R.x != R.y

    assert R.x*R.x == R.x*R.x
    assert R.x*R.x != R.x*R.y

    assert R.x*R.y == R.x*R.y
    assert R.x*R.y != R.x*R.z

    assert R.x*R.y == R.y*R.x
    assert R.x*R.y != R.x*R.z

def test_formula():
    assert R.x + R.y == R.x + R.y
    assert R.x + R.y != R.x + R.z

    assert R.x + R.y == R.y + R.x
    assert R.x + R.y != R.y + R.z

    assert R.x*(R.y+R.z) == R.x*R.y + R.x*R.z
    assert R.x*(R.y+R.z) != R.x*R.y + R.x*R.w

def test_addsub():
    assert I + I == I
    assert I + R.x != I

    assert R.x + R.x == R.x
    assert R.x + R.y != R.x

    assert I - I == O
    assert I - R.x != O

    assert (R.x + R.y) - R.y == R.x
    assert (R.x + R.y) - R.z != R.x

def test_droptype():
    assert drop_type(FIRST) == FIRST
    assert drop_type(NONE) == NONE
    assert drop_type(42) == VALUE
    assert drop_type(None) == VALUE

def test_drop():
    assert C.id._drop == FIRST
    assert C.id(FIRST)._drop == FIRST
    assert C.id(NONE)._drop == NONE
    assert C.id(42)._drop == 42
    assert C.id(None)._drop is None

def test_drop_multi():
    assert C.id1.to_term().drop(FIRST)._drop == FIRST
    assert C.id1.to_term().drop(NONE)._drop == NONE
    assert C.id1.to_term().drop(42)._drop == (42,)
    assert C.id1.to_term().drop(None)._drop == (None,)

    assert (C.id1*C.id2)._drop == FIRST
    assert (C.id1('A')*C.id2(1))._drop == ('A', 1)
    assert (C.id1('A')*C.id2(None))._drop == ('A', None)
    assert (C.id1*C.id2).drop('A', 1)._drop == ('A', 1)
    assert (C.id1*C.id2).drop(NONE)._drop == NONE
    assert (C.id1*C.id2).drop('A', None)._drop == ('A', None)
    assert Term(C.id1, C.id2)._drop == FIRST
    assert Term(C.id1, C.id2, drop=NONE)._drop == NONE
    assert Term(C.id1, C.id2, drop=('A', 1))._drop == ('A', 1)
    assert Term(C.id1, C.id2, drop=['A', 1])._drop == ('A', 1)

    assert (R.x1*R.x2)._drop == NONE
    assert (R.x*C.id)._drop == NONE
    assert (C.id1('A')*C.id2)._drop == FIRST
    assert (C.id1(FIRST)*C.id2(NONE))._drop == FIRST
    assert Term(C.id1('A'), C.id2(1), drop=FIRST)._drop == FIRST
    assert Term(C.id1('A'), C.id2, drop=NONE)._drop == NONE
    assert (C.id1('A')*C.id2(None))._drop == ('A', None)

    assert (C.id1*C.id2*C.id3)._drop == FIRST
    assert (C.id1('A')*C.id2(1)*C.id3)._drop == FIRST
    assert (C.id1('A')*C.id2(1)*C.id3(1.5))._drop == ('A', 1, 1.5)
    assert (C.id1('A')*C.id2(1)*C.id3(None))._drop == ('A', 1, None)
    assert (C.id1*C.id2*C.id3).drop('A', 1, 1.5)._drop == ('A', 1, 1.5)



import jax
import jax.lax as lax
from jax.scipy.special import gammaln
import jax.numpy as np
import jax.numpy.linalg as la
from jax.tree_util import (
    tree_flatten, tree_leaves, tree_map, tree_reduce, tree_structure
)
from jax.interpreters.xla import DeviceArray
import numpy as np0
import scipy.sparse as sp
import pandas as pd
from operator import and_, add

from .formula import (
    design_matrices, parse_item, parse_tuple, parse_list, ensure_formula, Categ
)
from .tools import block_inverse, chainer, maybe_diag, atleast_2d
from .summary import param_table

##
## constants
##

# numbers
eps = 1e-7

##
## canned models
##

def sigmoid(x):
    return 1/(1+np.exp(-x))

def log(x):
    return np.log(np.maximum(eps, x))

# link functions
links = {
    'ident': lambda x: x,
    'exp': np.exp,
    'logit': sigmoid
}

# loss functions
def binary_loss(yh, y):
    return y*log(yh) + (1-y)*log(1-yh)

def poisson_loss(yh, y):
    return y*log(yh) - yh

def negbin_loss(r, yh, y):
    return gammaln(r+y) - gammaln(r) + r*log(r) + y*log(yh) - (r+y)*log(r+yh)

def lstsq_loss(yh, y):
    return -(y-yh)**2

losses = {
    'binary': lambda p, yh, y: binary_loss(yh, y),
    'poisson': lambda p, yh, y: poisson_loss(yh, y),
    'negbin': lambda p, yh, y: negbin_loss(np.exp(p['lr']), yh, y),
    'lstsq': lambda p, yh, y: lstsq_loss(yh, y)
}

# modifiers
def zero_inflate(like0, clip_like=20.0, key='lpzero'):
    def like(p, yh, y):
        pzero = sigmoid(p[key])
        clike = np.clip(like0(p, yh, y), a_max=clip_like)
        like = pzero*(y==0) + (1-pzero)*np.exp(clike)
        return log(like)
    return like

##
## batching it, pytree style
##

class DataLoader:
    def __init__(self, data, batch_size=None):
        if type(data) is pd.DataFrame:
            data = data.to_dict('series')
        self.data = tree_map(
            lambda x: np.array(x) if type(x) is not np.ndarray else x, data
        )
        shapes = [d.shape[0] for d in tree_leaves(self.data)]
        self.data_size = shapes[0] # should all be the same size
        self.batch_size = batch_size

    def __call__(self, batch_size=None):
        yield from self.iterate(batch_size)

    def __iter__(self):
        yield from self.iterate()

    def iterate(self, batch_size=None):
        batch_size = batch_size if batch_size is not None else self.batch_size
        num_batches = max(1, self.data_size // batch_size)
        round_size = batch_size*num_batches
        for i in range(0, round_size, batch_size):
            yield tree_map(lambda d: d[i:i+batch_size], self.data)

class OneLoader:
    def __init__(self, data, batch_size=None):
        self.data = data

    def __iter__(self):
        yield from self.iterate()

    def iterate(self):
        yield self.data

##
## block matrices
##

def chunks(v, n):
    return [v[i:i+n] for i in range(0, len(v), n)]

def block_matrix(tree, dims, size):
    blocks = chunks(tree_leaves(tree), size)
    mat = np.block([
        [atleast_2d(x, axis=int(d>0)) for d, x in zip(dims, row)]
        for row in blocks
    ])
    return mat

def block_unpack(mat, tree, sizes):
    part = np.cumsum(np.array(sizes))
    block = lambda x, axis: tree.unflatten(np.split(x, part, axis=axis)[:-1])
    tree = tree_map(lambda x: block(x, 1), block(mat, 0))
    return tree

# we need par to know the inner shape
def tree_matfun(fun, mat, par):
    # get param configuration
    par_flat, par_tree = tree_flatten(par)
    par_sizs = [np.size(p) for p in par_flat]
    par_dims = [np.ndim(p) for p in par_flat]
    K = len(par_flat)

    # invert hessian for stderrs
    tmat = block_matrix(mat, par_dims, K)
    fmat = fun(tmat)
    fout = block_unpack(fmat, par_tree, par_sizs)

    return fout

def tree_batch_reduce(batch_fun, loader, agg_fun=np.add):
    total = None
    for b, batch in enumerate(loader):
        f_batch = batch_fun(batch)
        if total is None:
            total = f_batch
        else:
            total = tree_map(agg_fun, total, f_batch)
    return total

def tree_outer(tree):
    return tree_map(lambda x: tree_map(lambda y: x.T @ y, tree), tree)

def tree_outer_flat(tree):
    tree1, vec = dict_popoff(tree, 'hdfe')
    leaves = [atleast_2d(l) for l in tree_leaves(tree1)]
    mat = np.hstack(leaves)
    A = mat.T @ mat
    B = mat.T @ vec
    C = vec.T @ mat
    d = np.sum(vec*vec, axis=0)
    return A, B, C, d

def dict_popoff(d, s):
    if s in d:
        return {k: v for k, v in d.items() if k != s}, d[s]
    else:
        return d, None

def tree_fisher(vg_fun, params, loader):
    # accumulate outer product
    fish = tree_batch_reduce(
        lambda b: tree_outer(vg_fun(params, b)), loader
    )

    # invert fisher matrix
    sigma = tree_matfun(la.inv, fish, params)

    return sigma

def diag_fisher(vg_fun, params, loader):
    # compute hessian inverse by block
    A, B, C, d = tree_batch_reduce(
        lambda b: tree_outer_flat(vg_fun(params, b)), loader
    )
    psig, hsig = block_inverse(A, B, C, d, inv=la.inv)

    # unpack into tree
    par0, _ = dict_popoff(params, 'hdfe')
    par0_flat, par0_tree = tree_flatten(par0)
    par0_sizs = [np.size(p) for p in par0_flat]
    sigma = block_unpack(psig, par0_tree, par0_sizs)
    sigma['hdfe'] = hsig

    return sigma

# just get mean and var vectors
def flatten_output(beta, sigma):
    beta_real = beta['real']
    beta_categ = np.hstack([
        beta['categ'][c] for c in beta['categ']
    ])

    sigma_real = maybe_diag(sigma['real']['real'])
    sigma_categ = np.hstack([
        maybe_diag(sigma['categ'][c]['categ'][c]) for c in sigma['categ']
    ])

    beta_vec = np.hstack([beta_real, beta_categ])
    sigma_vec = np.hstack([sigma_real, sigma_categ])

    return beta_vec, sigma_vec

##
## optimizers
##

def adam(vg_fun, loader, params0, epochs=10, eta=0.01, gamma=0.9, disp=None):
    # parameter info
    params = tree_map(np.array, params0)

    # track rms gradient
    grms = tree_map(np.zeros_like, params)

    # do training
    for ep in range(epochs):
        # epoch stats
        agg_loss, agg_batch = 0.0, 0

        # iterate over batches
        for b, batch in enumerate(loader):
            # compute gradients
            loss, grad = vg_fun(params, batch)

            lnan = np.isnan(loss)
            gnan = tree_reduce(
                and_, tree_map(lambda g: np.isnan(g).any(), grad)
            )

            if lnan or gnan:
                print('Encountered nans!')
                return params, None

            grms = tree_map(
                lambda r, g: gamma*r + (1-gamma)*g**2, grms, grad
            )
            params = tree_map(
                lambda p, g, r: p + eta*g/np.sqrt(r+eps), params, grad, grms
            )

            # compute statistics
            agg_loss += loss
            agg_batch += 1

        # display stats
        avg_loss = agg_loss/agg_batch

        # display output
        if disp is not None:
            disp(ep, avg_loss, params)

    return params

##
## estimation
##

# maximum likelihood using jax - this expects a mean log likelihood
def maxlike(
    model=None, params=None, data=None, stderr=False, optim=adam, backend='gpu',
    **kwargs
):
    # get model gradients
    vg_fun = jax.jit(jax.value_and_grad(model), backend=backend)

    # simple non-batched loader
    loader = OneLoader(data)

    # maximize likelihood
    params1 = optim(vg_fun, loader, params, **kwargs)

    if not stderr:
        return params1, None

    # get model hessian
    h_fun = jax.jit(jax.hessian(model), backend=backend)

    # compute standard errors
    hess = h_fun(params, data)
    fish = tree_matfun(inv_fun, hess, params)
    omega = tree_map(lambda x: -x, fish)

    return params1, omega

# maximum likelihood using jax - this expects a mean log likelihood
# the assumes the data is batchable, which usually means panel-like
# a toplevel hdfe variable is treated special-like
def maxlike_panel(
    model=None, params=None, data=None, vg_fun=None, stderr=True, optim=adam,
    batch_size=8192, batch_stderr=8192, backend='gpu', **kwargs
):
    # compute gradient for optim
    vg_fun = jax.jit(jax.value_and_grad(model), backend=backend)

    # set up batching
    loader = DataLoader(data, batch_size)

    # maximize likelihood
    params1 = optim(vg_fun, loader, params, **kwargs)

    # just point estimates
    if not stderr:
        return params1, None

    # get vectorized gradient
    gv_fun = jax.jit(jax.vmap(jax.grad(model), (None, 0), 0), backend=backend)

    # batching for stderr
    hload = loader(batch_stderr)

    # compute standard errors
    if 'hdfe' in params:
        sigma = diag_fisher(gv_fun, params1, loader)
    else:
        sigma = tree_fisher(gv_fun, params1, loader)

    return params1, sigma

# make a glm model and compile
def glm_model(link, loss, hdfe=None, drop='first'):
    if type(link) is str:
        link = links[link]
    if type(loss) is str:
        loss = losses[loss]

    # evaluator
    def model(par, dat):
        ydat, xdat, cdat = dat['ydat'], dat['xdat'], dat['cdat']
        real, categ = par['real'], par['categ']
        if hdfe is not None:
            categ[hdfe] = par.pop('hdfe')
        linear = xdat @ real
        for i, c in enumerate(categ):
            cidx = cdat.T[i] # needed for vmap to work
            if drop == 'first':
                linear += np.where(cidx > 0, categ[c][cidx-1], 0.0)
            else:
                linear += categ[c][cidx]
        pred = link(linear)
        like = loss(par, pred, ydat)
        return np.mean(like)

    return model

# default glm specification
def glm(
    y=None, x=None, formula=None, hdfe=None, data=None, extra={}, model=None,
    link=None, loss=None, extern=None, stderr=True, drop='first', display=True,
    epochs=None, per=None, output='table', **kwargs
):
    # convert to formula system
    y, x = ensure_formula(x=x, y=y, formula=formula)

    # add in hdfe if needed
    if hdfe is not None:
        c_hdfe = parse_tuple(hdfe, convert=Categ)
        x += c_hdfe
        hdfe = c_hdfe.name()

    # construct design matrices
    y_vec, y_name, x_mat, x_names, c_mat, c_names0 = design_matrices(
        y=y, x=x, data=data, method='ordinal', extern=extern, flatten=False
    )

    # accumulate all names
    if drop == 'first':
        c_names0 = {k: v[1:] for k, v in c_names0.items()}
    c_names = chainer(c_names0.values())
    names = x_names + c_names

    # get data shape
    N = len(y_vec)
    Kx = len(x_names)
    Kc = len(c_names0)
    Kl = [len(c) for c in c_names0.values()]

    # choose number of epochs
    epochs = max(1, 2_000_000 // N) if epochs is None else epochs
    per = max(1, epochs // 5) if per is None else per

    # compile model if needed
    if model is None:
        model = glm_model(link, loss, hdfe=hdfe, drop=drop)

    # displayer
    def disp(e, l, p):
        real, categ = p['real'], p['categ']
        if hdfe is not None:
            categ = categ.copy()
            categ[hdfe] = p['hdfe']
        mcats = np.array([np.mean(c) for c in categ.values()])
        if e % per == 0:
            print(f'[{e:3d}] {l:.5f}: {np.mean(real):.5f} {np.mean(mcats):.5f}')
    disp1 = disp if display else None

    # organize data and initial params
    dat = {'ydat': y_vec, 'xdat': x_mat, 'cdat': c_mat}
    pcateg = {c.name(): np.zeros(s) for c, s in zip(c_names0, Kl)}
    params = {'real': np.zeros(Kx), 'categ': pcateg, **extra}
    if hdfe is not None:
        params['hdfe'] = params['categ'].pop(hdfe)

    # estimate model
    beta, sigma = maxlike_panel(
        model=model, params=params, data=dat, stderr=stderr, disp=disp1,
        epochs=epochs, **kwargs
    )

    if hdfe is not None:
        beta['categ'][hdfe] = beta.pop('hdfe')
        if stderr:
            sigma['categ'][hdfe] = {'categ': {hdfe: sigma.pop('hdfe')}}

    if output == 'table':
        beta_vec, sigma_vec = flatten_output(beta, sigma)
        return param_table(beta_vec, y_name, names, sigma=sigma_vec)
    elif output == 'dict':
        return {
            'beta': beta,
            'sigma': sigma,
        }

# logit regression
def logit(y=None, x=None, data=None, **kwargs):
    return glm(y=y, x=x, data=data, link='logit', loss='binary', **kwargs)

# poisson regression
def poisson(y=None, x=None, data=None, **kwargs):
    return glm(y=y, x=x, data=data, link='exp', loss='poisson', **kwargs)

# zero inflated poisson regression
def zinf_poisson(y=None, x=None, data=None, clip_like=20.0, **kwargs):
    return glm(
        y=y, x=x, data=data, link='exp',
        loss=zero_inflate(losses['poisson'], clip_like=clip_like),
        extra={'lpzero': 0.0}, **kwargs
    )

# negative binomial regression
def negbin(y=None, x=None, data=None, **kwargs):
    return glm(
        y=y, x=x, data=data, link='exp', loss='negbin', extra={'lr': 0.0},
        **kwargs
    )

# zero inflated poisson regression
def zinf_negbin(y=None, x=None, data=None, clip_like=20.0, **kwargs):
    return glm(
        y=y, x=x, data=data, link='exp',
        loss=zero_inflate(losses['negbin'], clip_like=clip_like),
        extra={'lpzero': 0.0, 'lr': 0.0}, **kwargs
    )

# ordinary least squares (just for kicks)
def ols_loss(p, yh, y):
    lsigma2 = p['lsigma2']
    sigma2 = np.exp(lsigma2)
    like = -lsigma2 + lstsq_loss(yh, y)/sigma2
    return like

def gols(y=None, x=None, data=None, **kwargs):
    return glm(
        y=y, x=x, data=data, link='ident', loss=ols_loss,
        extra={'lsigma2': 0.0}, **kwargs
    )



##
## ols test suite
##

import re
import pytest
import numpy as np
import pandas as pd
import statsmodels.formula.api as smf
import fastreg as fr
from fastreg import I, R, C, NONE

# standard formulas

y, y0 = R.y, R.y0
x0 = I + R.x1 + R.x2
x = x0 + C.id1 + C.id2

# helper functions

def statsmodels_rename(s):
    if s == 'Intercept':
        return 'I'
    elif ret := re.match(r'C\(([^\)]+)\)\[T\.([^\]]+)\]', s):
        return '='.join(ret.groups())
    else:
        return s

def statsmodels_frame(res):
    df = pd.DataFrame({
        'coeff': res.params,
        'stderr': np.sqrt(np.diagonal(res.cov_params()))
    })
    df = df.rename_axis(res.model.endog_names, axis=1)
    df = df.rename(statsmodels_rename, axis=0)
    return df

def statsmodels_isclose(fit_fr, fit_sm):
    ret_fr = fit_fr[['coeff', 'stderr']].sort_index()
    ret_sm = statsmodels_frame(fit_sm).sort_index()
    return np.isclose(ret_fr, ret_sm).all()

def frame_valid(df):
    return ~np.isnan(df).any().any()

# tests

def test_ols_real(data):
    ret = fr.ols(y=y0, x=x0, data=data)

    assert frame_valid(ret)

def test_ols_categ(data):
    ret = fr.ols(y=y, x=x, data=data)

    assert frame_valid(ret)

def test_ols_real_statsmodels(data):
    fit_fr = fr.ols(y=y0, x=x0, data=data)
    fit_sm = smf.ols('y0 ~ 1 + x1 + x2', data=data).fit()

    assert statsmodels_isclose(fit_fr, fit_sm)

def test_ols_categ_statsmodels(data):
    xp = R.x1 + R.x2 + C.id1(NONE) + C.id2
    fit0_fr = fr.ols(y=y, x=xp, data=data)
    fit0_sm = smf.ols(
        'y ~ 0 + x1 + x2 + C(id1) + C(id2)', data=data, eval_env=-1
    ).fit()

    fit1_fr = fr.ols(y=y, x=x, data=data)
    fit1_sm = smf.ols(
        'y ~ 1 + x1 + x2 + C(id1) + C(id2)', data=data, eval_env=-1
    ).fit()

    assert statsmodels_isclose(fit0_fr, fit0_sm)
    assert statsmodels_isclose(fit1_fr, fit1_sm)

def test_absorb(data):
    ret0 = fr.ols(y=y, x=x, data=data)
    ret1 = fr.ols(y=y, x=x, absorb=C.id1, data=data)
    ret2 = fr.ols(y=y, x=x, absorb=C.id2, data=data)

    elreg1 = r'(x1|x2|id2=.+)'
    cx01 = ret0.filter(regex=elreg1, axis=0)['coeff']
    cx11 = ret1.filter(regex=elreg1, axis=0)['coeff']

    elreg2 = r'(x1|x2|id1=.+)'
    cx02 = ret0.filter(regex=elreg2, axis=0)['coeff']
    cx22 = ret2.filter(regex=elreg2, axis=0)['coeff']

    assert np.isclose(cx01, cx11).all()
    assert np.isclose(cx02, cx22).all()

def test_hdfe(data):
    ret0 = fr.ols(y=y, x=x, data=data).sort_index()
    ret1 = fr.ols(y=y, x=x, hdfe=C.id1, data=data).sort_index()
    ret2 = fr.ols(y=y, x=x, hdfe=C.id2, data=data).sort_index()

    assert np.isclose(ret0, ret1).all()
    assert np.isclose(ret0, ret2).all()

@pytest.mark.parametrize('cov', [0, 1, 2, 3])
def test_ols_robust(data, cov):
    ret = fr.ols(y=y0, x=x0, data=data, stderr=f'hc{cov}')

    assert frame_valid(ret)

@pytest.mark.parametrize('cov', [0, 1, 2, 3])
def test_ols_robust_statsmodels(data, cov):
    fit_fr = fr.ols(y=y0, x=x0, data=data, stderr=f'hc{cov}')
    fit_sm = smf.ols('y0 ~ 1 + x1 + x2', data=data).fit(cov_type=f'HC{cov}')

    assert statsmodels_isclose(fit_fr, fit_sm)



##
## tools
##

import numpy as np
import pandas as pd
import numpy.linalg as la
import scipy.sparse as sp
from itertools import chain, accumulate
from functools import partial
from pandas._libs.hashtable import SIZE_HINT_LIMIT
from pandas.core.sorting import get_group_index

# general pointwise multiply
def multiply(a, b):
    return a.multiply(b) if sp.issparse(a) else a*b

# make a sparse matrix dense
def ensure_dense(x):
    return x.toarray() if sp.issparse(x) else x

# make a row vector into a matrix, maybe
def atleast_2d(x, axis=0):
    if x.ndim < 2:
        if axis == 0:
            return x.reshape((-1, 1))
        else:
            return x.reshape((1, -1))
    else:
        return x

# return vector or matrix diagonal
def maybe_diag(x):
    if x.ndim == 1:
        return x
    else:
        return x.diagonal()

# handles empty data
def vstack(v, format='csr'):
    v = [x for x in v if x is not None]
    if len(v) == 0:
        return None
    elif any([sp.issparse(x) for x in v]):
        return sp.vstack(v, format=format)
    else:
        return np.vstack(v)

# allows None's and handles empty data
def hstack(v, format='csr'):
    v = [x for x in v if x is not None]
    if len(v) == 0:
        return None
    elif any([sp.issparse(x) for x in v]):
        return sp.hstack(v, format=format)
    else:
        return np.hstack(v)

def get_dtype(*v):
    tarr = [x.dtype for x in v if hasattr(x, 'dtype')]
    tsca = [type(x) for x in v if not hasattr(x, 'dtype')]
    return np.lib.index_tricks.find_common_type(tarr, tsca)

# splice two 1d arrays based on boolean condition
def splice(cond, x1, x2):
    dtype = get_dtype(x1, x2)
    x = np.empty_like(cond, dtype=dtype)
    x[cond] = x1
    x[~cond] = x2
    return x

# concat lists
def chainer(v):
    return list(chain.from_iterable(v))

# factorize like pd.factorize but with 2d arrays
def factorize_2d(a):
    # column compressor
    size_hint = min(a.shape[0], SIZE_HINT_LIMIT)
    def f(vals):
        codes, labels = pd.factorize(vals, sort=True, size_hint=size_hint)
        return codes.astype("i8", copy=False), labels, len(labels)

    # get grouped ids
    vals = (col for col in a.T)
    codes, labels, shape = map(list, zip(*map(f, vals)))
    gids = get_group_index(codes, shape, sort=False, xnull=False)

    # get unique ids
    gcode, glabs = pd.factorize(gids, sort=True)
    coords = np.unravel_index(glabs, shape)
    ulabs = np.vstack([lab[crd] for crd, lab in zip(coords, labels)]).T

    return gcode, ulabs

# split list on boolean condition
def categorize(func, seq):
    true, false = [], []
    for item in seq:
        if func(item):
            true.append(item)
        else:
            false.append(item)
    return true, false

# handles sparse too (dense based on statsmodels version)
def group_sums(x, codes):
    if sp.issparse(x):
        _, K = x.shape
        x = x.tocsc()
        C = max(codes) + 1
        idx = [(x.indptr[i], x.indptr[i+1]) for i in range(K)]
        return np.vstack([
            np.bincount(
                codes[x.indices[i:j]], weights=x.data[i:j], minlength=C
            ) for i, j in idx
        ]).T
    elif np.ndim(x) == 1:
        return np.bincount(codes, weights=x)
    else:
        _, K = x.shape
        return np.vstack([
            np.bincount(codes, weights=x[:, j]) for j in range(K)
        ]).T

# sparsity handled by group_sums
def group_means(x, codes):
    if np.ndim(x) == 1:
        return group_sums(x, codes)/np.bincount(codes)
    else:
        return group_sums(x, codes)/np.bincount(codes)[:, None]

# block inversion with diagonal d
def block_inverse(A, B, C, d, inv=la.inv):
    d1 = 1/d
    d1l, d1r = d1[:, None], d1[None, :]
    A1 = inv(A - (B*d1r) @ C)
    Ai = A1
    di = d1 + np.sum((d1l*C)*(A1 @ (B*d1r)).T, axis=1)
    return Ai, di

# do isnan on general ndarrays
def fillna(x, v=0):
    return np.where(np.isnan(x), v, x)

# try to handle dense and various sparse formats
def valid_rows(x):
    fmt = x.format if sp.issparse(x) else None
    if fmt is None:
        null = ~pd.isnull(x)
        return null if x.ndim == 1 else null.all(axis=1)
    elif fmt == 'csr':
        N, _ = x.shape
        nidx, = np.nonzero(np.isnan(x.data))
        rows = np.unique(np.digitize(nidx, x.indptr)-1)
        null = np.isin(np.arange(N), rows)
    else:
        print(f'valid_rows: unsupported format "{fmt}"')
    return ~null

def all_valid(*mats):
    return np.vstack([m for m in mats if m is not None]).all(axis=0)

# list based cumsum
def cumsum(x):
    return list(accumulate(x))

# split by sizes rather than boundaries
def split_size(x, s):
    b = cumsum(s)
    return np.split(x, b)

# expects compact positive integers, except for -1 being null
# drops -1 always and 0 if drop (maps into all zeros)
def onehot_encode(indx, drop=True):
    base = 1 if drop else 0
    sel = indx >= base
    rows = np.flatnonzero(sel)
    cols = indx[sel] - base
    data = np.ones_like(rows)
    N, K = len(indx), cols.max() + 1
    smat = sp.csr_matrix(
        (data, (rows, cols)), shape=(N, K)
    )
    cats = np.arange(base, K+base)
    return smat, cats

##
## function tools
##

# decorator with optional flags
def decorator(decor0):
    def decor1(func=None, *args, **kwargs):
        if func is None:
            def decor2(func1):
                return decor0(func1, *args, **kwargs)
            return decor2
        else:
            return decor0(func, *args, **kwargs)
    return decor1

def func_name(func, anon='f'):
    fname = func.__name__
    if anon is not None and fname == '<lambda>':
        fname = anon
    return fname

def func_args(name, *args, **kwargs):
    astr = ','.join([f'{a}' for a in args])
    kstr = ','.join([f'{k}={v}' for k, v in kwargs.items()])
    sig = '|'.join(filter(len, [astr, kstr]))
    return f'{name}({sig})'

def func_disp(func, name=None):
    name = func_name(func) if name is None else name
    return partial(func_args, name)

##
## semi-defunct tools
##

def category_indices_numeric(vals, return_labels=False):
    # get unique indices
    labs, indx = np.unique(vals, axis=0, return_inverse=True)

    # patch in nans as -1
    nan_loc = np.flatnonzero(np.isnan(labs).any(axis=-1))
    if len(nan_loc) > 0:
        labs1 = np.delete(labs, nan_loc, axis=0)
        indx1 = np.where(np.isin(indx, nan_loc), -1, indx)
        _, indx2 = np.unique(indx1, return_inverse=True)
        indx, labs = indx2-1, labs1

    if return_labels:
        return indx, list(map(tuple, labs))
    else:
        return indx



##
## linear regressions
##

import re
import numpy as np
import pandas as pd
import scipy.sparse as sp
from functools import reduce
from operator import mul, and_

from .tools import (
    hstack, multiply, ensure_dense, group_sums, group_means, chainer,
    block_inverse, valid_rows
)
from .formula import (
    category_indices, design_matrices, parse_tuple, parse_list, ensure_formula,
    Categ
)
from .summary import param_table

def ols(
    y=None, x=None, formula=None, data=None, absorb=None, cluster=None, hdfe=None,
    stderr=True, extern=None, output='table'
):
    # convert to formula system
    y, x = ensure_formula(x=x, y=y, formula=formula)

    # track valid rows
    valid = np.ones(len(data), dtype=bool)

    # use hdfe trick
    if hdfe is not None:
        h_trm = parse_tuple(hdfe, convert=Categ)
        x = (x - h_trm) + h_trm # pop to end

    # don't include absorb
    if absorb is not None:
        a_trm = parse_list(absorb, convert=Categ)
        a_mat = a_trm.raw(data, extern=extern)
        valid &= reduce(and_, (valid_rows(a) for a in a_mat))
        x -= a_trm
        if cluster is None:
            cluster = reduce(mul, a_trm)

    # cluster=False is only to override auto on absorb
    if cluster is False:
        cluster = None

    # fetch cluster var
    if cluster is not None:
        s_trm = parse_tuple(cluster, convert=Categ)
        s_mat = s_trm.raw(data, extern=extern)
        valid &= valid_rows(s_mat)

    # make design matrices
    y_vec, y_name, x0_mat, x0_names, c_mat, c_names0, valid = design_matrices(
        y=y, x=x, formula=formula, data=data, valid0=valid, extern=extern,
        flatten=False, validate=True
    )

    # drop final invalid from cluster and absorb
    if absorb is not None:
        a_mat = [a[valid] for a in a_mat]
    if cluster is not None:
        s_mat = s_mat[valid]

    # combine x variables
    x_mat = hstack([x0_mat, c_mat])
    c_names = chainer(c_names0.values())
    x_names = x0_names + c_names

    # data shape
    N = len(data)
    K = len(x_names)

    # use absorption
    if absorb is not None:
        x_mat = ensure_dense(x_mat)
        y_vec, x_mat, keep = absorb_categorical(y_vec, x_mat, a_mat)
        if cluster is not None:
            s_mat = s_mat[keep, :]

    # linalg tool select
    if sp.issparse(x_mat):
        inv = sp.linalg.inv
        solve = sp.linalg.spsolve
    else:
        inv = np.linalg.inv
        solve = np.linalg.solve

    # find point estimates
    xpx = x_mat.T @ x_mat
    xpy = x_mat.T @ y_vec
    beta = solve(xpx, xpy)

    # just the point estimates
    if output == 'point':
        return beta

    # find residuals
    y_hat = x_mat @ beta
    e_hat = y_vec - y_hat

    # just the point estimate
    if stderr is False:
        if output == 'table':
            return param_table(beta, y_name, x_names)
        elif output == 'dict':
            return {
                'beta': beta,
                'y_name': y_name,
                'x_names': x_names,
                'y_hat': y_hat,
                'e_hat': e_hat,
            }

    # find inv(xpx) somehow
    if hdfe is not None:
        Kh = len(c_names0[h_trm])
        xh_mat, ch_mat = ensure_dense(x_mat[:, :-Kh]), x_mat[:, -Kh:]
        ixr, ixc = block_outer_inverse(xh_mat, ch_mat)
    else:
        ixpx = ensure_dense(inv(xpx))

    # compute classical sigma hat
    if stderr is True or hdfe is not None:
        s2 = (e_hat @ e_hat)/(N-K)

    # compute Xe moment
    if cluster is not None or type(stderr) is str:
        xe_mat = multiply(x_mat, e_hat[:, None])

    # find standard errors
    if cluster is not None:
        xe2 = error_sums(xe_mat, s_mat)
        sigma = ixpx @ xe2 @ ixpx
    elif hdfe is not None:
        sigma = s2*ixr, s2*ixc
    elif stderr is True:
        sigma = s2*ixpx
    elif type(stderr) is str:
        hcret = re.match(r'hc([0-3])', stderr.lower())
        if hcret is None:
            raise Exception(f'Unrecognized covariance type {stderr}')
        hc, = map(int, hcret.groups())
        sigma = hcn_stderr(hc, x_mat, xe_mat, ixpx)

    # return requested
    if output == 'table':
        return param_table(beta, y_name, x_names, sigma=sigma)
    elif output == 'dict':
        return {
            'beta': beta,
            'sigma': sigma,
            'y_name': y_name,
            'x_names': x_names,
            'y_hat': y_hat,
            'e_hat': e_hat,
        }

##
## standard errors
##

# inv(X.T @ X) when X = [X D] and D is sparse and diagonal
def block_outer_inverse(X, D):
    A = X.T @ X
    B = X.T @ D
    C = D.T @ X
    d = D.power(2).sum(axis=0).getA1()
    return block_inverse(A, B, C, d)

# from cameron and miller
def error_sums(xe, c):
    codes, _ = category_indices(c, dropna=True)
    xeg = group_sums(xe, codes)
    xe2 = xeg.T @ xeg
    return xe2

# handles hc in (0, 1, 2, 3)
def hcn_stderr(hc, x, xe, ixpx):
    N, K = x.shape
    if hc < 2:
        xeh = xe
    else:
        xq = x @ np.linalg.cholesky(ixpx)
        hii = np.sum(xq**2, axis=1)
        hinv = 1/(1-hii)
        if hc == 2:
            hinv = np.sqrt(hinv)
        xeh = multiply(xe, hinv[:, None])
    sigma = ixpx @ (xeh.T @ xeh) @ ixpx
    if hc == 1:
        sigma *= N/(N-K)
    return sigma

##
## absorption
##

# will absorb null (-1) values together
def absorb_categorical(y, x, abs):
    N, K = x.shape

    # copy so as not to destroy
    y = y.copy()
    x = x.copy()

    # store original means
    avg_y0 = np.mean(y)
    avg_x0 = np.mean(x, axis=0)

    # track whether to drop
    keep = np.ones(N, dtype=bool)

    # do this iteratively to reduce data loss
    for a in abs:
        # create class groups
        codes, _ = category_indices(a, dropna=True)

        # perform differencing on y
        avg_y = group_means(y, codes)
        y -= avg_y[codes]

        # perform differencing on x
        avg_x = group_means(x, codes)
        x -= avg_x[codes, :]

        # detect singletons
        multi = np.bincount(codes) > 1
        keep &= multi[codes]

    # recenter means
    y += avg_y0
    x += avg_x0[None, :]

    # drop invalid
    y = y[keep]
    x = x[keep, :]

    return y, x, keep



import numpy as np
import pandas as pd

from . import linear
from .formula import I, R, C

# true parameters
params0 = {
    'I': 0.1,
    'x1': 0.3,
    'x2': 0.6,
    'id1': 1.0,
    'id2': 1.0,
    'sigma': 1.0,
    'pz': 0.2,
    'alpha': 0.3
}

# poisson dampening
pfact = 100

# default specification
default_x = I + R('x1') + R('x2') + C('id1') + C('id2')

# good negbin in terms of mean and overdispersion (var = m + alpha*m^2)
def rand_negbin(mean, alpha, size=None, state=np.random):
    return state.negative_binomial(1/alpha, 1/(1+alpha*mean), size=size)

def dataset(
    N=1_000_000, K1=10, K2=100, models=['linear'], letter=True, params=params0,
    seed=89320432,
):
    if type(models) is str:
        models = [models]

    # init random
    st = np.random.RandomState(seed)

    # core regressors
    df = pd.DataFrame({'x1': st.randn(N), 'x2': st.randn(N)})

    # predictors
    df['yhat0'] = params['I'] + params['x1']*df['x1'] + params['x2']*df['x2']
    df['yhat'] = df['yhat0']

    if K1 is not None:
        df['id1'] = st.randint(K1, size=N)
        df['yhat'] += params['id1']*df['id1']/K1

    if K2 is not None:
        df['id2'] = st.randint(K2, size=N)
        df['yhat'] += params['id2']*df['id2']/K2

    # linear
    if 'linear' in models:
        df['y0'] = df['yhat0'] + params['sigma']*st.randn(N)
        df['y'] = df['yhat'] + params['sigma']*st.randn(N)

    # logit
    if 'logit' in models:
        df['Eb0'] = 1/(1+np.exp(-df['yhat0']))
        df['Eb'] = 1/(1+np.exp(-df['yhat']))
        df['b0'] = (st.randn(N) < df['Eb0']).astype(np.int)
        df['b'] = (st.randn(N) < df['Eb']).astype(np.int)

    # poisson
    if 'poisson' in models:
        df['Ep0'] = np.exp(df['yhat0'])
        df['Ep'] = np.exp(df['yhat'])
        df['p0'] = st.poisson(df['Ep0'])
        df['p'] = st.poisson(df['Ep'])

    # zero-inflated poisson
    if 'zinf_poisson' in models:
        df['pz0'] = np.where(st.rand(N) < params['pz'], 0, df['p0'])
        df['pz'] = np.where(st.rand(N) < params['pz'], 0, df['p'])

    # negative binomial
    if 'negbin' in models:
        df['nb0'] = rand_negbin(df['Ep0'], params['alpha'], state=st)
        df['nb'] = rand_negbin(df['Ep'], params['alpha'], state=st)

    # zero-inflated poisson
    if 'zinf_negbin' in models:
        df['nbz0'] = np.where(st.rand(N) < params['pz'], 0, df['nb0'])
        df['nbz'] = np.where(st.rand(N) < params['pz'], 0, df['nb'])

    if letter and K1 is not None:
        df['id1'] = df['id1'].map(lambda x: chr(65+x))

    return df

def dataset_compare(N=10_000_000, K=100):
    K1, K2 = N // K, K

    id1 = np.random.randint(K1, size=N)
    id2 = np.random.randint(K2, size=N)

    x1 = 5*np.cos(id1) + 5*np.sin(id2) + np.random.randn(N)
    x2 = np.cos(id1) + np.sin(id2) + np.random.randn(N)

    y= 3*x1 + 5*x2 + np.cos(id1) + np.cos(id2)**2 + np.random.randn(N)

    return pd.DataFrame({
        'y': y, 'x1': x1, 'x2': x2, 'id1': id1,  'id2': id2
    })

def plot_coeff(beta, params=params0):
    import matplotlib.pyplot as plt

    coeff = pd.DataFrame({
        'id2': np.arange(len(beta)),
        'beta1': beta
    })
    coeff['beta0'] = params['id2']*coeff['id2']/pfact
    coeff['beta0'] -= coeff['beta0'].mean()
    coeff['beta1'] -= coeff['beta1'].mean()

    # inferred ranges
    bmin = coeff[['beta0', 'beta1']].min().min()
    bmax = coeff[['beta0', 'beta1']].max().max()
    bvec = np.linspace(bmin, bmax, 100)

    # plot estimates
    fig, ax = plt.subplots(figsize=(6, 5))
    coeff.plot.scatter(x='beta0', y='beta1', ax=ax, alpha=0.5)
    ax.plot(bvec, bvec, c='r', linewidth=1, zorder=1)

    ax.set_xlabel('$\\beta_0$')
    ax.set_ylabel('$\\beta_1$')

def test_ols(data, y='y', x=default_x, plot=False, **kwargs):
    table = linear.ols(y=y, x=x, data=data, **kwargs)

    if plot:
        plot_coeff(table['coeff'].filter(regex='id2'))

    return table

def test_glm(data, estim='poisson', y='p', x=default_x, plot=False, **kwargs):
    from . import general

    if type(estim) is str:
        estim = getattr(general, estim)

    table = estim(y=y, x=x, data=data, **kwargs)

    if plot:
        plot_coeff(table['coeff'].filter(regex='id2'))

    return table



import pytest
import fastreg as fr

seed = 2385198437

@pytest.fixture(params=[10_000, 100_000])
def data(request):
    N = request.param
    return fr.dataset(N=N, K1=5, K2=20, models='linear', seed=seed)



##
## results summary
##

import numpy as np
import pandas as pd
from scipy.stats.distributions import norm

from .tools import maybe_diag

##
## constants
##

z95 = norm.ppf(0.975)

##
## param summary
##

def param_table(beta, y_name, x_names, sigma=None):
    # basic frame
    frame = pd.DataFrame({
        'coeff': beta,
    }, index=x_names)
    frame = frame.rename_axis(y_name, axis=1)

    # handle sigma cases
    if sigma is None:
        return frame
    elif type(sigma) is tuple:
        sigr, sigc = sigma
        stderr = np.sqrt(np.hstack([maybe_diag(sigr), sigc]))
    else:
        stderr = np.sqrt(maybe_diag(sigma))

    # confidence interval
    low95 = beta - z95*stderr
    high95 = beta + z95*stderr

    # p-value
    zscore = beta/stderr
    pvalue = 2*(1-norm.cdf(np.abs(zscore)))

    # stderr stats
    frame = frame.assign(
        stderr=stderr, low95=low95, high95=high95, pvalue=pvalue
    )

    return frame



##
## formula operations
##

import re
import sys
import numpy as np
import pandas as pd
import scipy.sparse as sp

from functools import reduce
from operator import add
from enum import Enum

from .meta import MetaFactor, MetaTerm, MetaFormula, MetaReal, MetaCateg
from .tools import (
    categorize, hstack, chainer, decorator, func_disp, valid_rows, split_size,
    atleast_2d, fillna, all_valid, splice, factorize_2d, onehot_encode
)

##
## tools
##

def is_categorical(ft):
    if isinstance(ft, MetaFactor):
        return isinstance(ft, MetaCateg)
    elif isinstance(ft, MetaTerm):
        return any(is_categorical(t) for t in ft)

def ensure_tuple(t):
    if type(t) is tuple:
        return t
    elif type(t) is list:
        return tuple(t)
    else:
        return t,

def robust_eval(data, expr, extern=None):
    # short circuit
    if expr is None:
        return None

    # extract values
    if type(expr) is str:
        vals = data.eval(expr, engine='python', local_dict=extern)
    elif callable(expr):
        vals = expr(data)
    else:
        vals = expr

    # ensure array
    if type(vals) is pd.Series:
        return vals.values
    elif type(vals) is np.ndarray:
        return vals
    else:
        return np.full(len(data), vals)

# this works with __add__ overload
def sum0(items):
    return reduce(add, items)

# category drop options
class Drop(Enum):
    NONE = 0
    FIRST = 1
    VALUE = 2

##
## categoricals
##

# make labels
def swizzle(ks, vs):
    return ','.join([f'{k}={v}' for k, v in zip(ks, vs)])

# ordinally encode interaction terms (tuple-like things)
# null data is returned with a -1 index if not dropna
def category_indices(vals, dropna=False, return_labels=False):
    # also accept single vectors
    vals = atleast_2d(vals)

    # track valid rows
    valid = valid_rows(vals)
    vals1 = vals[valid]

    # find unique rows
    uni_indx, uni_vals = factorize_2d(vals1)

    # patch in valid data
    if dropna or valid.all():
        uni_ind1 = uni_indx
    else:
        uni_ind1 = splice(valid, uni_indx, -1)

    # return requested
    if return_labels:
        return uni_ind1, uni_vals, valid
    else:
        return uni_ind1, valid

def zero_category(vals, labs, lab0):
    try:
        idx0 = labs.index(lab0)
    except ValueError:
        raise Exception(f'drop label not found: {lab0}')

    # promote to zero index
    labs = labs.copy()
    labs = [labs.pop(idx0), *labs]

    # shift value indices
    vals = np.where(vals == idx0, -1, vals)
    vals = np.where(vals < idx0, vals+1, vals)

    return vals, labs

# encode categories as one-hot matrix
def encode_categorical(vals, names, method='sparse', drop=Drop.FIRST):
    # reindex categoricals jointly
    cats_val, cats_lab, valid = category_indices(vals, return_labels=True)
    cats_lab = [swizzle(names, l) for l in cats_lab]

    # if ordinal no labels are dropped
    if method == 'ordinal':
        cats_enc, cats_use = cats_val.reshape(-1, 1), cats_lab
    elif method == 'sparse':
        dtype = drop_type(drop)
        if dtype == Drop.VALUE:
            sdrop = swizzle(names, drop)
            cats_val, cats_lab = zero_category(cats_val, cats_lab, sdrop)
            drop_first = True
        else:
            drop_first = dtype == Drop.FIRST
        cats_enc, cats_all = onehot_encode(cats_val, drop=drop_first)
        cats_use = [cats_lab[i] for i in cats_all]

    return cats_enc, cats_use, valid

# subset data allowing for missing chunks
def drop_invalid(valid, *mats, warn=True):
    V, N = np.sum(valid), len(valid)
    if V == 0:
        raise Exception('all rows contain null data')
    elif V < N:
        if warn:
            print(f'dropping {N-V}/{N} null data points')
        mats = [
            m[valid] if m is not None else None for m in mats
        ]
    return *mats,

# remove unused categories (sparse `mat` requires `labels`)
def prune_categories(mat, labels=None, method='sparse', warn=True):
    # get positive count cats
    if method == 'sparse':
        vcats = np.ravel((mat!=0).sum(axis=0)) > 0
        Kt = [len(ls) for ls in labels.values()]
        tvalid = split_size(vcats, Kt)
        P, K = np.sum(vcats), len(vcats)
    elif method == 'ordinal':
        vcats = [np.bincount(cm) > 0 for cm in mat.T]
        P = sum([np.sum(vc) for vc in vcats])
        K = sum([len(vc) for vc in vcats])

    # prune if needed
    if P < K:
        if warn:
            print(f'pruning {K-P}/{K} unused categories')

        # modify data matrix
        if method == 'sparse':
            mat = mat[:, vcats]
            labels = {
                t: [l for l, v in zip(ls, vs) if v]
                for (t, ls), vs in zip(labels.items(), tvalid)
            }
        elif method == 'ordinal':
            mat = np.vstack([category_indices(cm) for cm in mat.T]).T

    return mat, labels

def drop_type(d):
    return d if d in (Drop.FIRST, Drop.NONE) else Drop.VALUE

def drop_repr(d):
    t = drop_type(d)
    if t == Drop.VALUE:
        return ','.join([str(x) for x in d])
    else:
        _, s = str(t).split('.')
        return s

# aggregate drop values from list of Factors to Term
def consensus_drop(drops):
    if len(drops) == 0:
        return None
    udrop, *rdrop = set([drop_type(d) for d in drops])
    if len(rdrop) > 0:
        return Drop.FIRST
    elif udrop == Drop.VALUE:
        return tuple(drops)
    else:
        return udrop

##
## formula structure
##

class AccessorType(type):
    def __getattr__(cls, expr):
        return cls(expr)

class Factor(MetaFactor, metaclass=AccessorType):
    def __init__(self, expr, name=None):
        if type(expr) is str:
            self._expr = expr
            self._name = expr if name is None else name
        elif callable(expr):
            self._expr = expr
            self._name = name
        elif type(expr) is pd.Series:
            self._expr = expr.values
            self._name = expr.name
        else:
            self._expr = np.array(expr)
            self._name = name

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if isinstance(other, MetaFactor):
            return str(self) == str(other)
        else:
            return False

    def __repr__(self, **kwargs):
        return self._name

    def __add__(self, other):
        if isinstance(other, (MetaFactor, MetaTerm)):
            return Formula(self, other)
        elif isinstance(other, MetaFormula):
            return Formula(self, *other)
        else:
            raise TypeError(f'Not a valid addition: {other}')

    def __sub__(self, other):
        return Formula(self) - other

    def __mul__(self, other):
        if isinstance(other, MetaFactor):
            return Term(self, other)
        elif isinstance(other, MetaTerm):
            return Term(self, *other)
        elif isinstance(other, MetaFormula):
            return Formula(*[Term(self, *t) for t in other])
        else:
            raise TypeError(f'Not a valid multiplier: {other}')

    def __call__(self, *args, **kwargs):
        cls = type(self)
        return cls(self._expr, *args, **kwargs)

    def to_term(self, **kwargs):
        return Term(self, **kwargs)

    def name(self):
        return self._name

    def raw(self, data, extern=None):
        return robust_eval(data, self._expr, extern=extern)

    def eval(self, data, **kwargs):
        return self.to_term().eval(data, **kwargs)

class Term(MetaTerm):
    def __init__(self, *facts, drop=None):
        self._facts = facts
        if drop is None:
            if all(is_categorical(f) for f in facts):
                self._drop = consensus_drop([f._drop for f in facts])
            else:
                self._drop = Drop.NONE
        else:
            if drop_type(drop) == Drop.VALUE:
                self._drop = tuple(drop)
            else:
                self._drop = drop

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if isinstance(other, MetaTerm):
            return set(self) == set(other)
        elif isinstance(other, MetaFactor):
            return set(self) == {other}
        else:
            return False

    def __repr__(self):
        if len(self) == 0:
            return 'I'
        else:
            ds = f'|{drop_repr(self._drop)}' if self._drop != Drop.NONE else ''
            return '*'.join([f.__repr__(drop=False) for f in self]) + ds

    def __iter__(self):
        return iter(self._facts)

    def __len__(self):
        return len(self._facts)

    def __add__(self, other):
        if isinstance(other, (MetaFactor, MetaTerm)):
            return Formula(self, other)
        elif isinstance(other, MetaFormula):
            return Formula(self, *other)
        else:
            raise TypeError(f'Not a valid addition: {other}')

    def __sub__(self, other):
        return Formula(self) - other

    def __mul__(self, other):
        if isinstance(other, MetaFactor):
            return Term(*self, other)
        elif isinstance(other, MetaTerm):
            return Term(*self, *other)
        elif isinstance(other, MetaFormula):
            return Formula(*[Term(*self, *t) for t in other])
        else:
            raise TypeError(f'Not a valid multiplier: {other}')

    def name(self):
        return '*'.join([f.name() for f in self])

    def drop(self, *drop):
        if len(drop) == 1:
            drop0, = drop
            if drop_type(drop0) == Drop.VALUE:
                self._drop = drop
            else:
                self._drop = drop0
        else:
            self._drop = drop
        return self

    def raw(self, data, extern=None):
        return np.vstack([f.raw(data, extern=extern) for f in self]).T

    def eval(self, data, method='sparse', extern=None):
        # zero length is identity
        if len(self) == 0:
            N = len(data)
            return np.ones((N, 1)), np.ones(N, dtype=bool), ['I']

        # separate pure real and categorical
        categ, reals = categorize(is_categorical, self)
        categ, reals = Term(*categ), Term(*reals)

        # handle categorical
        if len(categ) > 0:
            categ_mat = categ.raw(data, extern=extern)
            categ_nam = [c.name() for c in categ]
            categ_value, categ_label, categ_valid = encode_categorical(
                categ_mat, categ_nam, method=method, drop=self._drop
            )

        # handle reals
        if len(reals) > 0:
            reals_mat = reals.raw(data, extern=extern)
            reals_value = reals_mat.prod(axis=1).reshape(-1, 1)
            reals_label = reals.name()
            reals_valid = valid_rows(reals_value)

        # combine results
        if len(categ) == 0:
            return reals_value, reals_valid, [reals_label]
        elif len(reals) == 0:
            return categ_value, categ_valid, categ_label
        else:
            # filling nulls with 0 keeps sparse the same
            term_value = categ_value.multiply(fillna(reals_value, v=0))
            term_label = [f'({l})*{reals_label}' for l in categ_label]
            term_valid = categ_valid & reals_valid
            return term_value, term_valid, term_label

class Formula(MetaFormula):
    def __init__(self, *terms):
        self._terms = tuple(dict.fromkeys(
            t if isinstance(t, MetaTerm) else Term(t) for t in terms
        )) # order preserving unique

    def __eq__(self, other):
        self_set = {frozenset(t) for t in self}
        if isinstance(other, MetaFormula):
            return self_set == {frozenset(t) for t in other}
        if isinstance(other, MetaTerm):
            return self_set == {frozenset(other)}
        elif isinstance(other, MetaFactor):
            return self_set == {frozenset({other})}
        else:
            return False

    def __repr__(self):
        return ' + '.join(str(t) for t in self)

    def __iter__(self):
        return iter(self._terms)

    def __len__(self):
        return len(self._terms)

    def __add__(self, other):
        if isinstance(other, (MetaFactor, MetaTerm)):
            return Formula(*self, other)
        elif isinstance(other, MetaFormula):
            return Formula(*self, *other)
        else:
            raise TypeError(f'Not a valid addition: {other}')

    def __sub__(self, other):
        if isinstance(other, MetaFactor):
            other = Term(other)
        if isinstance(other, MetaTerm):
            return Formula(*[
                t for t in self if t != other
            ])
        if isinstance(other, MetaFormula):
            return Formula(*[
                t for t in self if t not in other
            ])
        else:
            raise TypeError(f'Not a valid subtraction: {other}')

    def __mul__(self, other):
        if isinstance(other, MetaFactor):
            return Formula(*[Term(*t, other) for t in self])
        elif isinstance(other, MetaTerm):
            return Formula(*[Term(*t, *other) for t in self])
        elif isinstance(other, MetaFormula):
            return Formula(*chainer([
                [Term(*t1, *t2) for t1 in self] for t2 in other
            ]))
        else:
            raise TypeError(f'Not a valid multiplier: {other}')

    def raw(self, data, extern=None):
        return [t.raw(data, extern=extern) for t in self]

    def eval(self, data, method='sparse', extern=None):
        # split by all real or not
        categ, reals = categorize(is_categorical, self)

        # handle categories
        if len(categ) > 0:
            categ_value, categ_valid, categ_label = zip(*[
                t.eval(data, method=method, extern=extern) for t in categ
            ])
            categ_value = hstack(categ_value)
            categ_valid = np.vstack(categ_valid).all(axis=0)
        else:
            categ_value, categ_valid, categ_label = None, None, []

        # combine labels
        categ_label = {t: ls for t, ls in zip(categ, categ_label)}

        # handle reals
        if len(reals) > 0:
            reals_value, reals_valid, reals_label = zip(*[
                t.eval(data, extern=extern) for t in reals
            ])
            reals_value = hstack(reals_value)
            reals_label = chainer(reals_label)
            reals_valid = np.vstack(reals_valid).all(axis=0)
        else:
            reals_value, reals_valid, reals_label = None, None, []

        # return separately
        return (
            reals_value, reals_valid, reals_label,
            categ_value, categ_valid, categ_label
        )

##
## column types
##

class Real(MetaReal, Factor):
    def __repr__(self, **kwargs):
        return f'R({self.name()})'

class Categ(MetaCateg, Factor):
    def __init__(self, expr, drop=Drop.FIRST, **kwargs):
        super().__init__(expr, **kwargs)
        self._drop = drop

    def __repr__(self, drop=True, **kwargs):
        nm = self.name()
        if drop and self._drop != Drop.NONE:
            ds = drop_repr(self._drop)
            return f'C({nm}|{ds})'
        else:
            return f'C({nm})'

    def drop(self, drop):
        self._drop = drop
        return self

# custom columns — class interface
# raw (mandatory): an ndarray of the values
# name (recommended): what gets displayed in the regression table
# __repr__ (optional): what gets displayed on print [default to C/R(name)]

class Demean(Real):
    def __init__(self, expr, cond=None, name=None):
        args = '' if cond is None else f'|{cond}'
        name = expr if name is None else name
        super().__init__(expr, name=f'{name}-μ{args}')
        self._cond = cond

    def raw(self, data, extern=None):
        vals = super().raw(data, extern=extern)
        if self._cond is None:
            means = np.mean(vals)
        else:
            cond = robust_eval(data, self._cond, extern=extern)
            datf = pd.DataFrame({'vals': vals, 'cond': cond})
            cmean = datf.groupby('cond')['vals'].mean().rename('mean')
            datf = datf.join(cmean, on='cond')
            means = datf['mean'].values
        return vals - means

class Binned(Categ):
    def __init__(self, expr, bins=10, drop=Drop.FIRST, labels=False, name=None):
        nb = bins if type(bins) is int else len(bins)
        name = expr if name is None else name
        super().__init__(expr, drop=drop, name=f'{name}:bin{nb}')
        self._bins = bins
        self._labels = None if labels else False

    def raw(self, data, extern=None):
        vals = super().raw(data, extern=extern)
        bins = pd.cut(vals, self._bins, labels=self._labels)
        return bins

# shorthand for drop=NONE

class Categ0(Categ):
    def __init__(self, expr, *args, drop=None, **kwargs):
        super().__init__(expr, *args, drop=Drop.NONE, **kwargs)

class Binned0(Binned):
    def __init__(self, expr, *args, drop=None, **kwargs):
        super().__init__(expr, *args, drop=Drop.NONE, **kwargs)

# custom columns — functional interface

class Custom:
    def __init__(
        self, func, name=None, categ=False, base=Real, eval_args=0, frame=False
    ):
        self._base = Categ if categ else base
        self._func = func
        self._name = name if callable(name) else func_disp(func, name=name)
        self._eval = ensure_tuple(eval_args)
        self._frame = frame

    def __getattr__(self, key):
        if key.startswith('_'):
            return super().__getattr__(key)
        else:
            return self(key)

    def __call__(self, *args, **kwargs):
        if self._frame:
            evaler = lambda data: self._func(data, *args, **kwargs)
        else:
            def evaler(data):
                args1 = [
                    robust_eval(data, e) if i in self._eval else e
                    for i, e in enumerate(args)
                ]
                return self._func(*args1, **kwargs)
        name = self._name(*args, **kwargs)
        return self._base(evaler, name=name)

@decorator
def factor(func, *args, **kwargs):
    return Custom(func, *args, **kwargs)

# shortcuts
O = Formula()
I = Term()
R = Real
C = Categ
D = Demean
B = Binned
C0 = Categ0
B0 = Binned0

##
## conversion
##

# lookup table
FTYPES = {
    'C': Categ,
    'I': Real,
}

def parse_factor(fact):
    ret = re.match(r'(C|I)\(([^\)]+)\)', fact.code)
    if ret is not None:
        pre, name = ret.groups()
        return FTYPES[pre](name)
    else:
        return Real(fact.code)

def parse_term(term):
    return Term(*[parse_factor(f) for f in term.factors])

# this can only handle treatment coding, but that's required for sparsity
def parse_formula(form):
    try:
        from patsy.desc import ModelDesc
    except:
        print('Please install patsy for formula parsing')
        return

    # use patsy for formula parse
    desc = ModelDesc.from_formula(form)
    lhs, rhs = desc.lhs_termlist, desc.rhs_termlist

    # check for invalid y
    if len(lhs) > 1:
        raise Exception('Must have single factor y term')

    # convert to string lists
    y_terms = parse_factor(lhs[0].factors[0]) if len(lhs) > 0 else None
    x_terms = Formula(*[parse_term(t) for t in rhs])

    return y_terms, x_terms

def parse_item(i, convert=Real):
    if isinstance(i, MetaFactor):
        return i
    else:
        return convert(i)

def parse_tuple(t, convert=Real):
    if isinstance(t, MetaTerm):
        return t
    else:
        if type(t) not in (tuple, list):
            t = t,
        return Term(*[
            parse_item(i, convert=convert) for i in t
        ])

def parse_list(l, convert=Real):
    if isinstance(l, MetaFormula):
        return l
    else:
        if type(l) not in (tuple, list):
            l = l,
        return Formula(*[
            parse_tuple(t, convert=convert) for t in l
        ])

##
## design interface
##

def ensure_formula(y=None, x=None, formula=None):
    if formula is not None:
        y, x = parse_formula(formula)
    else:
        if isinstance(y, (MetaTerm, MetaFormula)):
            raise Exception(f'LHS variable must be a single factor. Instead got: y = {y}')
        else:
            y = parse_item(y) if y is not None else None
        x = parse_list(x)
    return y, x

def design_matrix(
    x=None, formula=None, data=None, method='sparse', dropna=True, prune=True,
    warn=True, extern=None, valid0=None, flatten=True, validate=False
):
    _, x = ensure_formula(x=x, formula=formula)

    # evaluate x variables
    x_mat, x_val, x_names, c_mat, c_val, c_labels = x.eval(
        data, method=method, extern=extern
    )

    # aggregate valid info for data
    valid = all_valid(valid0, x_val, c_val)

    # drop null values if requested
    if dropna:
        x_mat, c_mat = drop_invalid(
            valid, x_mat, c_mat, warn=warn
        )

    # prune empty categories if requested
    if prune and c_mat is not None:
        c_mat, c_labels = prune_categories(
            c_mat, c_labels, method=method, warn=warn
        )

    # combine real and categorical?
    if flatten:
        f_mat = hstack([x_mat, c_mat])
        f_names = x_names + chainer(c_labels.values())
        ret = f_mat, f_names
    else:
        ret = x_mat, x_names, c_mat, c_labels

    # return valid mask?
    if validate:
        return *ret, valid
    else:
        return ret

def design_matrices(
    y=None, x=None, formula=None, data=None, dropna=True, extern=None,
    valid0=None, validate=False, **kwargs
):
    # parse into pythonic formula system
    y, x = ensure_formula(x=x, y=y, formula=formula)
    if y is None:
        raise Exception('Use design_matrix for formulas without an LHS')

    # get y data
    y_vec, y_name = y.raw(data, extern=extern), y.name()
    y_val = valid_rows(y_vec)

    # get valid x data
    x_val0 = all_valid(valid0, y_val)
    *x_ret, valid = design_matrix(
        x=x, data=data, dropna=dropna, extern=extern, valid0=x_val0,
        validate=True, **kwargs
    )

    # drop invalid y
    if dropna:
        y_vec, = drop_invalid(valid, y_vec)

    # return combined data
    ret = y_vec, y_name, *x_ret
    if validate:
        return *ret, valid
    else:
        return ret
